// handle all output
#include"IO.h"
using namespace std;

void IO::openfiles() {
    char tname[80];
    sprintf(tname,"thermodata.d%d.N%d.phi%.4f.T%.3f.attfrac%.5f",d,N,phi,Ttarg,attfrac);
    thermout.open(tname);
    char mname[80];
    sprintf(mname,"mindata.d%d.N%d.phi%.4f.T%.3f.attfrac%.5f",d,N,phi,Ttarg,attfrac);
    minout.open(mname);
    char vname[80];
    sprintf(vname,"vanHoveCorr.d%d.N%d.phi%.4f.T%.3f.attfrac%.5f",d,N,phi,Ttarg,attfrac);
    vhcout.open(vname);
    char xname[80];
    sprintf(xname,"coords.d%d.N%d.phi%.4f.T%.3f.attfrac%.5f",d,N,phi,Ttarg,attfrac);
    xout.open(xname);
    return;
}


// Close all files
void IO::closefiles() {
    thermout.close();
    minout.close();
    vhcout.close();
    xout.close();
    return;
}


// Write sigma, coordinates, and image flags to config-dump file
void IO::writecoords(int step) {
    int i, j;
    xout.precision(10);
    xout << step << " " << lo << " " << hi << endl;
    for (i = 0; i < N; i++) {
        xout << i << " " << sigma[i];
        for (j = 0; j < d; j++) xout << " " << RVF[3*d*i+j];
        for (j = 0; j < d; j++) xout << " " << image[d*i+j];
        xout << endl;
    }
    return;
}



// Write mass, sigma, epsilon, coords, velocities, image flags to restart file
void IO::writerestart(int step) {
    int i, j;
    char rname[80];
    sprintf(rname,"restart.d%d.N%d.phi%.4f.T%.3f.attfrac%.4f.step%d",d,N,phi,Ttarg,attfrac,step);
    resout.open(rname);
    resout.precision(10);
    resout << step << " " << lo << " " << hi << endl;
    for (i = 0; i < N; i++) {
        resout << i << " " << mass[i] << " " << sigma[i] << " " << epsilon[i];
        for (j = 0; j < 2*d; j++) resout << " " << RVF[3*d*i+j];
        for (j = 0; j < d; j++) resout << " " << image[d*i+j];
        resout << endl;
    }
    resout.close();
    return;
}



// Write initial coords, image flags
void IO::writeRinit(int step) {
    int i, j;
    char inifname[80];
    sprintf(inifname,"Rinit.d%d.N%d.phi%.4f.T%.3f.attfrac%.5f",d,N,phi,Ttarg,attfrac);
    initout.open(inifname);
    initout.precision(10);
    initout << step << " " << lo << " " << hi << endl;
    for (i = 0; i < N; i++) {
        initout << i;
        for (j = 0; j < d; j++) initout << " " << RVF[3*d*i+j];
        for (j = 0; j < d; j++) initout << " " << image[d*i+j];
        initout << endl;
    }
    initout.close();
    return;
}


// Output thermodynamic data
void IO::outputthermo(int step) {
    if (!step) thermout << "Step T E Epair P Z phi namsd ngp f_ov" << endl;
    thermout << step << " " << T << " " << E << " " << EP << " " << P << " " << Z << " " << phi << " " << namsd << " " << ngp << " " << f_ov << endl;
    return;
}


// Output energy-minimization data
void IO::outputmin(int step, int iteration, double fnorm, double fracDE) {
    if (step == 0 && iteration == 0) minout << "Step Iteration T Epair Pressure Z Z_noratt f_norm/(d*N) EP_del/EP Rattlers Floaters namsd ngp f_ov time" << endl;
    if (step >= 0 && iteration >= 0) {
        minout << step << " " << iteration << " " << T << " " << EP << " " << P << " " << Z << " " << Znoratt << " ";
        minout << fnorm/(d*N) << " " << fracDE << " " << rattlers << " " << floaters << " " << msd << " " << ngp << " " << f_ov << " " << min_actual_time << endl;
    }
    return;
}


// Output van Hove correlation function
void IO::outputVHC(double *vhc, int step) {
    int i;
    vhcout << step << endl;
    for (i = 0; i < 2500; i++) vhcout << (i+.5)*.01 << " " << vhc[i] << endl;
    return;
}
