// initialize.cpp -- initialize RVF, ALU, m, sigma, epsilon, etc.
#include<cmath>
#include"omp.h"
#include"cellmanagement.h"
#include"string.h"
#include<mm_malloc.h>
#define PI 3.1415926536
using namespace std;
extern double ALU[3];



// Intialize position, force, velocity, image, neighbor arrays
// Initialize mass and sigma distributions; see RevisitingPolyDisp.nb and CheckPolyDisp.nb
// Use P(σ) = Aσ^-d form from Ninarello17: this ensures different-size particles occupy equal volume fractions
// MVR = ratio of max/minimum particle volumes: code slows down only very slowly with increasing MVR
// _mm_malloc() makes the memory occupied by the given array --aligned--
double initRVFMSE(bool readresflag, const char *initresfname, const char *initRinitfname) {
    int i, j, k;
    double A, L, U;  // normalization prefactor, minimum and maximum particle diameters
//    double MVR = d;  // script-R in the paper
    double MVR = pow(1.0/.88,d);  // for polydisperse generalization of KA
    L = pow(MVR,-1.0/(2*d));        U = pow(MVR,1.0/(2*d));
    A = ((d-1.0)/d)/(pow(MVR,(1.0-d)/(2*d*d))-pow(MVR,(d-1.0)/(2*d*d)));
    ALU[0] = A;     ALU[1] = L;     ALU[2] = U;
    double trialm, ranr, prob;
    double totpartvol = 0;
    int initstep;
    double initlo, inithi;
    double vs[11] = {0, 1, PI/4, PI/6, PI*PI/32, PI*PI/60, pow(PI,3)/384, pow(PI,3)/840, pow(PI,4)/6144, pow(PI,4)/15120, pow(PI,5)/122880}; // volumes of d-dimensional unit hyperspheres
    
    RVF = (double*) _mm_malloc(3*d*N*sizeof(double),64);
    Rlast = (double*) _mm_malloc(d*N*sizeof(double),64);
    Rinit = (double*) _mm_malloc(d*N*sizeof(double),64);
    image = (short int*) _mm_malloc(d*N*sizeof(short int),16);
    initimg = (short int*) _mm_malloc(d*N*sizeof(short int),16);
    allneigh = (int*) _mm_malloc(N*maxneigh*sizeof(int),32);
    numneigh = (short int*) _mm_malloc(N*sizeof(short int),16);
    mass = (double*) _mm_malloc(N*sizeof(double),64);
    sigma = (double*) _mm_malloc(N*sizeof(double),64);
    epsilon = (double*) _mm_malloc(N*sizeof(double),64);
    sqrtepsilon = (double*) _mm_malloc(N*sizeof(double),64);
    Z_array = (int*) _mm_malloc(N*sizeof(int),64);
    
    ifstream resin, rinitin;
    
// If readresflag = false, initialize mass, sigma, and epsilon of all particles,
// give all particles random positions, zero velocities and forces,
// and set Rinit and Rlast to current positions
    srand(time(NULL));
    if (!readresflag) {
        step = 0;
        i = 0;
        do  {
            trialm = ALU[1] + ((ALU[2]-ALU[1])*rand())/RAND_MAX;
            ranr = (1.0*rand())/RAND_MAX;
            prob = pow(trialm/ALU[1],-d);
            if (ranr < prob) {
                sigma[i] = trialm;
                mass[i] = pow(sigma[i],d);  // Unit-diameter particles have unit mass
//                mass[i] = 1;   // Use to avoid getting "trivially" heterogeneous dynamics
                epsilon[i] = pow(sigma[i],d);   // Particles' "charges" are proportional to their volumes
                totpartvol += vs[d]*pow(sigma[i],d);
                i++;
            }
        } while (i < N);
// Set cell boundaries
        vol = totpartvol/phi;
        rho = N/vol;
        D = pow(vol,1.0/d);     Dinit = D;
        lo = -D/2;  hi = D/2;
// THEN initialize particle positions
        for (k = 0; k < d*N; k++) {
            RVF[(3*d)*(k/d) + k%d] = lo + (D*rand())/RAND_MAX;
            RVF[(3*d)*(k/d) + k%d + d] = 0;
            RVF[(3*d)*(k/d) + k%d + 2*d] = 0;
            Rlast[k] = RVF[(3*d)*(k/d) + k%d];
            Rinit[k] = RVF[(3*d)*(k/d) + k%d];
            image[k] = 0;   initimg[k] = 0;
        }
    }
// If readresflag = true, read all of the above in from user-specifed restart and Rinit files
    else {
        resin.open(initresfname);
        if (!resin.is_open()) { cout << "Failed to open restart-file " << initresfname << endl;    exit(1);   }
        resin >> step >> lo >> hi;
        D = hi - lo;
        vol = pow(D,d);
        rho = N/vol;
        for (i = 0; i < N; i++) {
            resin >> k >> mass[i] >> sigma[i] >> epsilon[i];
            for (j = 0; j < 2*d; j++) resin >> RVF[3*d*i+j];
            for (j = 0; j < d; j++) resin >> image[d*i+j];
            for (j = 0; j < d; j++) Rlast[d*i+j] = RVF[3*d*i+j];
            totpartvol += vs[d]*pow(sigma[i],d);
        }
        resin.close();
// If not using an Rinit file, set Rinit = R and initimg = image
        if (!strcmp(initRinitfname,"null")) {
            for (k = 0; k < d*N; k++) {
                Rinit[k] = RVF[(3*d)*(k/d) + k%d];
                initimg[k] = image[k];
            }
            Dinit = D;
        }
// If using an Rinit file, get Rinit and initimg from this file
        else if (strcmp(initRinitfname,"null")) {
            rinitin.open(initRinitfname);
            if (!rinitin.is_open()) { cout << "Failed to open Rinit-file " << initRinitfname << endl;    exit(1);   }
            rinitin >> initstep >> initlo >> inithi;
            Dinit = inithi - initlo;
            for (i = 0; i < N; i++) {
                rinitin >> k;
                for (j = 0; j < d; j++) rinitin >> Rinit[d*i+j];
                for (j = 0; j < d; j++) rinitin >> initimg[d*i+j];
            }
            rinitin.close();
        }
    }
    
// Set sqrtepsilon (speeds implementation of the Lorentz-Berthelot mixing rules)
    for (i = 0; i < N; i++) sqrtepsilon[i] = sqrt(epsilon[i]);
    
    return totpartvol;
}



// Deallocate all arrays at end of run
void dealloc() {
    _mm_free(RVF);      _mm_free(Rlast);    _mm_free(Rinit);
    _mm_free(image);    _mm_free(initimg);
    _mm_free(allneigh);     _mm_free(numneigh);
    _mm_free(mass);     _mm_free(sigma);    _mm_free(epsilon);
    _mm_free(sqrtepsilon);
    
    return;
}
